@@ -35,7 +35,8 @@ function(n, theta, alpha = 1)
     alpha <- alpha / sum(alpha)
 
     y <- matrix(0, n, ncol(theta))
-    ind <- as.numeric(cut(runif(n), c(0, cumsum(alpha)), include.lowest = TRUE))
+    ind <- as.numeric(cut(runif(n), c(0, cumsum(alpha)),
+                          include.lowest = TRUE))
     pos <- split(seq_len(n), ind)
     nms <- names(pos)
     for(i in seq_along(pos)) {
@@ -59,7 +60,7 @@ function(x, ...)
 rW <-
 function(n, kappa, d)
 {
-    .C("rW",
+    .C(C_rW,
        as.integer(n),
        as.double(kappa),
        as.integer(d),
